uniform sampler2D colorMap;
uniform sampler2D normalMap;
uniform sampler2D heightMap;
uniform sampler2D specularMap;

uniform int useNormalMap;
uniform int useHeightMap;
uniform int useSpecularMap;

uniform int useParallaxMapping;
uniform float displacement;

varying vec3 eyeVec;
varying vec3 _eyeVec;
varying vec3 lightVec;
varying vec3 _lightVec;
varying vec3 halfVec;
varying vec3 normalVec;

void main() {
  vec2 texCoord = gl_TexCoord[0].xy;
  float offsetHeight;

  vec3 normal = normalVec;
  if (useNormalMap == 1) {
    if (useParallaxMapping > 0 && useHeightMap == 1 && displacement > 0.0) {
	  // Parallax Mapping
	  if (useParallaxMapping == 1) {
	    vec2 offset = displacement * vec2(eyeVec); 
		offsetHeight = texture2D(heightMap, texCoord+offset).r;    
		if (offsetHeight < 0.96) {
		  offset *= (offsetHeight + texture2D(heightMap, texCoord).r) * 0.5;          
		}

		texCoord += vec2(offset);
	  }
	  // Steep Parallax Mapping
	  else if (useParallaxMapping > 1) {
	    float height = 1.0;
		int numSteps = 30;
		if (useParallaxMapping == 3) {
		  numSteps = int(1000.0 * displacement);
		}

		numSteps = int(mix(2*numSteps, numSteps, eyeVec.z));
		
		float stepLength = 1.0 / float(numSteps);
		vec2 delta = vec2(-_eyeVec.x, -_eyeVec.y) * displacement / (_eyeVec.z * float(numSteps));

		offsetHeight = texture2D(heightMap, texCoord).r;

		for (int i = 0; i < numSteps; i++) {
		  if (offsetHeight < height) {
			height -= stepLength;
			texCoord += delta;
			offsetHeight = texture2D(heightMap, texCoord).r;
		  } 
		  else {
			break;
		  }
		}
	  }	  
	}

	normal = normalize(texture2D(normalMap, texCoord).rgb*2.0 - 1.0);
  }

  vec3 ambient = vec3(gl_FrontMaterial.ambient) * vec3(gl_LightSource[1].ambient);
  vec3 color = ambient;  

  float NdotL = max(dot(normal, lightVec), 0.0);
    
  float selfShadow = 1.0;

  if (NdotL > 0.0) {
    if (useParallaxMapping == 3 && useHeightMap == 1 && displacement > 0.0) {
	  int numShadowSteps = int(mix(60, 5, lightVec.z));
	  float stepLength = 1.0 / float(numShadowSteps);
	  vec2 delta = vec2(_lightVec.x, _lightVec.y) * displacement / (_lightVec.z * float(numShadowSteps));

	  float height = offsetHeight + stepLength * 0.1;

	  vec2 offsetCoord = texCoord;

	  while ((offsetHeight < height) && (height < 1.0)) {
	    height += stepLength;
	    offsetCoord += delta;
	    offsetHeight = texture2D(heightMap, offsetCoord).r;
	  }

	  selfShadow = (offsetHeight < height) ? 1.0 : 0.5;
	}

    vec4 materialDiffuse = texture2D(colorMap, texCoord) * gl_Color;

	vec3 diffuse = vec3(materialDiffuse) * vec3(gl_LightSource[1].diffuse);
	vec3 specular = vec3(gl_FrontMaterial.specular) * vec3(gl_LightSource[1].specular); 

	if (useSpecularMap == 1) {
	  specular = vec3(texture2D(specularMap, texCoord));
	}

	float NdotHV = max(dot(normal, halfVec), 0.0);

	color += diffuse * NdotL;
	color += specular * pow(NdotHV, gl_FrontMaterial.shininess);
	color *= selfShadow;
  }  
  
  gl_FragColor = vec4(color, gl_Color.a);
}